/* 
 * Metronome session
 *
 * This file is part of GTick
 * 
 *
 * Copyright (c) 1999, Alex Roberts
 * Copyright (c) 2003, Roland Stigge <stigge@antcom.de>
 *
 * GTick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GTick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GTick; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include <config.h>

/* GNU headers */
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

/* GTK+ headers */
#include <gtk/gtk.h>

/* own headers */
#include "globals.h"
#include "metro.h"
#include "dsp.h"
#include "mixer.h"
#include "help.h"
#include "gtkutil.h"
#include "manualtimer.h"
#include "util.h"
#include "option.h"
#include "gtkoptions.h"

/* icons */
#include "icon32x32.xpm"
#include "icon48x48.xpm"
#include "icon64x64.xpm"

#define TIMER_DELAY 300
volatile int interrupted = 0; /* the caught signal will be stored here */

static state_data_t state_data[] = {
  { N_("Idle"),    N_("Start") },  /* STATE_IDLE    */
  { N_("Running"), N_("Stop") }    /* STATE_RUNNING */
};

/* handler for SIGTERM and SIGINT */
static void terminate_signal_callback(int sig) {
  interrupted = sig;
}

static void gtAdjVol(GtkWidget* widget _U_, gpointer data)
{
  metro_t* metro = (metro_t*) data;
  
  gtSetMixer(GTK_ADJUSTMENT(metro->vAdj)->value);
}

/*
 * Rate callback
 */
static void gtUpdateRate (GtkWidget* widget _U_, gpointer data)
{
  metro_t *metro = (metro_t *)data;
  
  metro->frequency = (GTK_ADJUSTMENT(metro->adjustment)->value) / 60.0;

  if (debug)
    g_print ("gtUpdateRate: rate=%f bpm\n", metro->frequency * 60.0);
}

/*
 * Change state
 */
static void gtSetState(metro_t* metro, int state) {
 
  if (state != metro->state) {
    
    if (metro->state == STATE_RUNNING && state == STATE_IDLE) {
      gtCloseDSP(metro);
    } else if (metro->state == STATE_IDLE && state == STATE_RUNNING) {
      /* Initialise /dev/dsp */ 
      if ((gtInitDSP(metro)) == -1) {
	gtCloseDSP(metro);
        GtkWidget* message = gtk_message_dialog_new(GTK_WINDOW(metro->window),
            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            _("Couldn't start metronome.\n"
	      "Please check if specified sound device\n"
	      "and sample file are accessible."));
        /* Destroy the dialog when the user responds to it */
        g_signal_connect_swapped(G_OBJECT(message),
                                 "response",
                                 G_CALLBACK(gtk_widget_destroy),
                                 G_OBJECT(message));
	gtk_widget_show(message);
	
	state = STATE_IDLE;
      }
    } else {
      fprintf(stderr, "Warning: Unhandled state change");
    }
    
    gtk_statusbar_pop(GTK_STATUSBAR(metro->statusbar),
		      metro->status_context);
    gtk_statusbar_push(GTK_STATUSBAR(metro->statusbar),
		       metro->status_context,
		       _(state_data[state].state));
    gtk_label_set_text(GTK_LABEL(metro->togglebutton_label),
		       _(state_data[state].toggle_label));
    
    metro->state = state;
  }
}

/*
 * Callback for toggle button (metronome on/off ...)
 */
static void gtToggle(GtkWidget* widget _U_, gpointer data) {
  metro_t* metro = (metro_t*) data;
  int new_state;
  
  if (metro->state == STATE_IDLE) {
    new_state = STATE_RUNNING;
  } else {
    new_state = STATE_IDLE;
  }

  gtSetState(metro, new_state);
}

/*
 * Quit program and destroy callback
 */
static void gtQuit(metro_t *metro, guint action _U_, GtkWidget* widget _U_)
{
  gtSetState(metro, STATE_IDLE);
  gtk_main_quit();
}

/*
 * periodically (e.g. each 300 milliseconds) from gtk main loop called function
 */
static gint timeout_callback(metro_t* metro) {
  if (interrupted) {
    switch(interrupted) {
    case SIGINT:
      if (debug)
	fprintf(stderr, "SIGINT (Ctrl-C ?) caught.\n");
      break;
    case SIGTERM:
      if (debug)
	fprintf(stderr, "SIGTERM caught.\n");
      break;
    default:
      fprintf(stderr, "Warning: Unknown signal caught.\n");
    }
    gtQuit(metro, 0, NULL);
  }

  return TRUE; /* call it again */
}
/*
 * Callback for meter mode
 */
static void gtMeter(GtkWidget *widget, gpointer data) {
  metro_t *metro = (metro_t*) data;

  if (widget == metro->btn0 ||
      widget == metro->btn24 ||
      widget == metro->btn34 ||
      widget == metro->btn44)
    gtk_widget_set_sensitive(metro->meterspinbutton, FALSE);
  
  if (widget == metro->btn0) 
    metro->meter = 1;
  else if (widget == metro->btn24)
    metro->meter = 2;
  else if (widget == metro->btn34)
    metro->meter = 3;
  else if (widget == metro->btn44)
    metro->meter = 4;
  else { // spinbutton activated or changed
    if (widget == metro->btnother) // other: spinbutton activated
      gtk_widget_set_sensitive(metro->meterspinbutton, TRUE);
    metro->meter = gtk_spin_button_get_value_as_int(
	GTK_SPIN_BUTTON(metro->meterspinbutton));
  }

  if (debug)
    g_print("new meter: %d\n", metro->meter);
}

/*
 * indirecting callback showing tapping_dialog_new()
 */
static void gtManualTimer(metro_t* metro,
                          guint action _U_,
			  GtkWidget* widget _U_)
{
  GtkWidget* dialog = tapping_dialog_new(metro->window, metro->adjustment);

  gtk_widget_show(dialog);
}

/*
 * indirecting callback showing gtk_options_dialog_new()
 */
static void gtProperties(metro_t* metro,
                         guint action _U_,
			 GtkWidget* widget _U_)
{
  GtkWidget* dialog = gtk_options_dialog_new(metro->window, metro->options);

  gtk_widget_show(dialog);
}

/*
 * option system callback for choosing the sample
 *
 * returns 0 on success, -1 otherwise
 */
static int set_sample(metro_t* metro, const char* samplename) {
  free(metro->options->samplename);
  metro->options->samplename = strdup(samplename);
  if (metro->state == STATE_RUNNING) {
    gtCloseDSP(metro);
    if (gtInitDSP(metro) == -1)
      return -1;
  }

  return 0;
}

/*
 * option system callback for spotting the sample
 */
static const char* get_sample(metro_t* metro) {
  return metro->options->samplename;
}

/*
 * returns the (main) menu bar for the specified window
 */
static GtkWidget* get_menubar(metro_t* metro) {
  GtkItemFactoryEntry menu_items[] = {
  /* path, accelerator, cb, cb parameter, item type, extra */
{_("/_Metronome"), NULL, NULL, 0, "<Branch>", NULL},
{_("/Metronome/_Manual Timer"), NULL, gtManualTimer, 1, NULL, NULL},
{_("/Metronome/"), NULL, NULL, 0, "<Separator>", NULL},
{_("/Metronome/_Quit"), "<control>Q", gtQuit, 1, "<StockItem>", GTK_STOCK_QUIT},

{_("/_Options"), NULL, NULL, 0, "<Branch>", NULL},
{_("/Options/_Properties"), NULL, gtProperties, 1, "<StockItem>", GTK_STOCK_PROPERTIES},

{_("/_Help"), NULL, NULL, 0, "<LastBranch>", NULL},
{_("/Help/_About"), NULL, gtAboutBox, 0, NULL, NULL},
{_("/Help/_License"), NULL, gtLicense, 0, NULL, NULL}
  };
  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  GtkAccelGroup* accel_group = gtk_accel_group_new();
  GtkItemFactory* item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR,
                                                      "<main>", accel_group);

  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, metro);
  gtk_window_add_accel_group(GTK_WINDOW(metro->window), accel_group);

  return gtk_item_factory_get_widget(item_factory, "<main>");
}

/*
 * returns a new metro object
 */
metro_t* metro_new(void) {
  metro_t* metro;
  GtkObject* adjustment;
  GList* icon_list = NULL;
  GtkWidget* windowvbox;
  GtkWidget* vbox;
  GtkWidget* metroframe;
  GtkWidget* metrovbox;
  GtkWidget* menu;
  GtkWidget* label;
  GtkWidget* table;
  GtkWidget* beatbox;
  GtkWidget* separator;
  GtkWidget* spinbutton;
  GtkWidget* hbox;
  GtkWidget* togglebutton;

  if (!(metro = g_malloc(sizeof(metro_t)))) {
    fprintf(stderr, "Error allocating memory for metro object.\n");
    exit(1);
  }
  if (debug)
    printf("Creating metro object at %p.\n", metro);
  
  metro->options = options_new();
  metro->dspfd = -1;
  metro->sound_event_id = G_MAXUINT;
  metro->fragment = NULL;
  metro->tickdata0 = NULL;
  metro->tickdata1 = NULL;
  metro->tickdata2 = NULL;
  metro->silence = NULL;

  option_register(&metro->options->option_list,
                  "SampleFilename",
		  (option_set_t) set_sample,
		  (option_get_t) get_sample,
		  (void*) metro);
  
  metro->state = STATE_IDLE; /* set default */
  
  metro->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  g_signal_connect_swapped(G_OBJECT(metro->window),
                           "destroy",
		           G_CALLBACK(gtQuit),
		           metro);

  gtk_window_set_title (GTK_WINDOW (metro->window), TITLE " " VERSION);

  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon32x32));
  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon48x48));
  icon_list = g_list_append(icon_list,
          (gpointer) gdk_pixbuf_new_from_xpm_data((const char**) icon64x64));
  gtk_window_set_icon_list(GTK_WINDOW(metro->window), icon_list);

  windowvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add(GTK_CONTAINER(metro->window), windowvbox);
  gtk_widget_show (windowvbox);

  menu = get_menubar(metro);
  gtk_box_pack_start(GTK_BOX(windowvbox), menu, FALSE, TRUE, 0);
  gtk_widget_show(menu);

  vbox = gtk_vbox_new(FALSE, 8);
  gtk_box_pack_start_defaults(GTK_BOX(windowvbox), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
  gtk_widget_show(vbox);

  metro->statusbar = gtk_statusbar_new();
  metro->status_context =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(metro->statusbar), "status");
  gtk_statusbar_push(GTK_STATUSBAR(metro->statusbar),
                     metro->status_context, 
		     _(state_data[metro->state].state));
  gtk_box_pack_end(GTK_BOX(windowvbox), metro->statusbar, FALSE, TRUE, 0);
  gtk_widget_show(metro->statusbar);

  /* Metronome UI Code */
  metroframe = gtk_frame_new(_("Settings"));
  gtk_box_pack_start_defaults(GTK_BOX(vbox), metroframe);
  gtk_widget_show (metroframe);

  metrovbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(metrovbox), 10);
  gtk_container_add(GTK_CONTAINER(metroframe), metrovbox);
  gtk_widget_show(metrovbox);

  table = gtk_table_new(2, 3, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(metrovbox), table);
  gtk_table_set_row_spacings(GTK_TABLE(table), 10);
  gtk_table_set_col_spacings(GTK_TABLE(table), 10);
  gtk_widget_show(table);
  
  label = gtk_label_new(_("Speed:"));
  gtk_table_attach(GTK_TABLE(table), label,
                   0, 1, 0, 1,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_widget_show (label);

  metro->adjustment = gtk_adjustment_new(75.0, MIN_BPM, MAX_BPM,
                                         1.0, 1.0, 1.0);
  gtk_signal_connect (GTK_OBJECT (metro->adjustment), "value_changed",
				  GTK_SIGNAL_FUNC (gtUpdateRate),
				  (metro_t *)metro);
		      
  gtUpdateRate (metro->window, (metro_t *)metro); /* Set initial rate */

  metro->scale = gtk_hscale_new (GTK_ADJUSTMENT (metro->adjustment));
  gtk_range_set_update_policy (GTK_RANGE (metro->scale), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_digits (GTK_SCALE (metro->scale), 1);
  gtk_scale_set_draw_value (GTK_SCALE (metro->scale), FALSE);
  gtk_table_attach(GTK_TABLE(table), metro->scale,
                   1, 2, 0, 1,
		   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show (metro->scale);

  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(metro->adjustment),
                                   1.0, /* Step */
				   1    /* number of decimals */);
  gtk_table_attach(GTK_TABLE(table), spinbutton,
                   2, 3, 0, 1,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show(spinbutton);

  label = gtk_label_new (_("Volume:"));
  gtk_table_attach(GTK_TABLE(table), label,
                   0, 1, 1, 2,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_widget_show (label);
	  
  metro->vAdj = gtk_adjustment_new (0.0, 0.0, 100.0, 5.0, 25.0, 0.0);
  gtk_signal_connect (GTK_OBJECT (metro->vAdj), "value_changed",
				  GTK_SIGNAL_FUNC (gtAdjVol),
				  (metro_t *)metro);	
  gtk_adjustment_set_value(GTK_ADJUSTMENT(metro->vAdj), 
			   (gfloat)(gtGetMixer()));
  metro->vol = gtk_hscale_new (GTK_ADJUSTMENT(metro->vAdj));
  gtk_range_set_update_policy (GTK_RANGE (metro->vol), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_digits (GTK_SCALE (metro->vol), 0);
  gtk_scale_set_draw_value (GTK_SCALE (metro->vol), FALSE);
  gtk_table_attach(GTK_TABLE(table), metro->vol,
                   1, 2, 1, 2,
		   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show (metro->vol);
  
  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(metro->vAdj),
                                   1.0, /* Step */
				   1    /* number of decimals */);
  gtk_table_attach(GTK_TABLE(table), spinbutton,
                   2, 3, 1, 2,
		   GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_show(spinbutton);
  
  /* Timing UI Code */
  
  metro->meter = 1; /* Default, 1 = None, 2 = 2/4, 3 = 3/4, 4 = 4/4 */

  beatbox = gtk_frame_new (_("Beat"));
  gtk_box_pack_start_defaults(GTK_BOX(metrovbox), beatbox);
  gtk_widget_show (beatbox);

  hbox = gtk_hbox_new (FALSE, 20);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
  gtk_container_add (GTK_CONTAINER (beatbox), hbox);
  gtk_widget_show (hbox);

  metro->btn0 = gtk_radio_button_new_with_label (NULL, _("Even"));
  gtk_box_pack_start (GTK_BOX (hbox), metro->btn0, TRUE, TRUE, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (metro->btn0), TRUE);
  gtk_widget_show (metro->btn0);

  gtk_signal_connect(GTK_OBJECT(metro->btn0), "pressed",
		     GTK_SIGNAL_FUNC(gtMeter),
		     metro);
  
  metro->btn24 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->btn0),
	"2/4");
  gtk_box_pack_start(GTK_BOX(hbox), metro->btn24, TRUE, TRUE, 0);
  gtk_widget_show (metro->btn24);
  gtk_signal_connect(GTK_OBJECT(metro->btn24), "pressed",
		     GTK_SIGNAL_FUNC(gtMeter),
		     metro);
  
  metro->btn34 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->btn24),
	"3/4");
  gtk_box_pack_start(GTK_BOX(hbox), metro->btn34, TRUE, TRUE, 0);
  gtk_widget_show (metro->btn34);
  gtk_signal_connect(GTK_OBJECT(metro->btn34), "pressed",
		     GTK_SIGNAL_FUNC(gtMeter),
		     metro);

  metro->btn44 =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->btn34),
	"4/4");
  gtk_box_pack_start(GTK_BOX(hbox), metro->btn44, TRUE, TRUE, 0);
  gtk_widget_show (metro->btn44);
  gtk_signal_connect(GTK_OBJECT(metro->btn44), "pressed",
		     GTK_SIGNAL_FUNC(gtMeter),
		     metro);

  metro->btnother =
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(metro->btn44),
	_("Other:"));
  gtk_box_pack_start(GTK_BOX(hbox), metro->btnother,
      TRUE, TRUE, 0);
  gtk_widget_show(metro->btnother);
  gtk_signal_connect(GTK_OBJECT(metro->btnother), "pressed",
      GTK_SIGNAL_FUNC(gtMeter), metro);
  
  adjustment = gtk_adjustment_new(5, 1, 99, 1, 4, 0);
  metro->meterspinbutton =
    gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 1, 0);
  gtk_widget_set_sensitive(metro->meterspinbutton, FALSE);
  gtk_widget_show(metro->meterspinbutton);
  gtk_box_pack_start(GTK_BOX(hbox), metro->meterspinbutton,
      TRUE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(metro->meterspinbutton), "value-changed",
      GTK_SIGNAL_FUNC(gtMeter), metro);

  /* General UI Code */
  separator = gtk_hseparator_new ();
  gtk_box_pack_start_defaults(GTK_BOX(windowvbox), separator);
  gtk_widget_show (separator);

  hbox = gtk_hbox_new (FALSE, 10);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
  gtk_box_pack_start (GTK_BOX(windowvbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show (hbox);

  togglebutton = gtk_button_new();
  gtk_signal_connect (GTK_OBJECT (togglebutton), "clicked",
				  GTK_SIGNAL_FUNC (gtToggle),
				  metro);
			   
  gtk_box_pack_start(GTK_BOX(hbox), togglebutton, TRUE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS (togglebutton, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (togglebutton);
  GTK_WIDGET_SET_FLAGS (togglebutton, GTK_CAN_FOCUS);
  gtk_widget_grab_focus (togglebutton);
  gtk_widget_show (togglebutton);

  metro->togglebutton_label =
    gtk_label_new(_(state_data[metro->state].toggle_label));
  gtk_misc_set_padding(GTK_MISC(metro->togglebutton_label), 64, 0);
  gtk_container_add(GTK_CONTAINER(togglebutton),
		    metro->togglebutton_label);
  gtk_widget_show(metro->togglebutton_label);

  /* set up additional handlers */
  gtk_timeout_add(TIMER_DELAY, (GtkFunction) timeout_callback, metro);
  signal(SIGINT, &terminate_signal_callback);
  signal(SIGTERM, &terminate_signal_callback);

  gtk_widget_show (metro->window);

  return metro;
}

void metro_delete(metro_t* metro) {
  options_delete(metro->options);
  if (debug)
    printf("Destroying metro object at %p.\n", metro);
  g_free(metro);
}
