/*
 * option.c: Options handling functions
 *
 * This file is part of GTick
 *
 * Copyright (c) 1999, Alex Roberts
 * Copyright (c) 2003, Roland Stigge <stigge@antcom.de>
 *
 * GTick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GTick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GTick; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

/* GNU headers */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* own headers */
#include "globals.h"
#include "option.h"

/*
 * returns new, empty options list
 */
option_list_t* option_list_new(void) {
  return NULL;
}

/*
 * destroys list, releasing all memory associated with it
 */
void option_list_delete(option_list_t* list) {
  option_list_t* temp;

  while (list) {
    temp = list;
    list = list->next;
    free(temp);
  }
}

/*
 * inserts new option entry to specified list* with:
 * 
 * name: name of the option
 * getter, setter: getter and setter functions for this option
 * object: argument to deliver to setter and getter
 *
 * returns 0 on success, -1 otherwise
 */
int option_register(option_list_t** list,
                    const char* name,
		    option_set_t setter,
		    option_get_t getter,
		    void* object)
{
  option_list_t* temp;

  if (!(temp = malloc(sizeof(option_list_t)))) {
    return -1;
  }

  temp->name = name;
  temp->setter = setter;
  temp->getter = getter;
  temp->object = object;
  temp->next = *list;
  *list = temp;

  return 0;
}

/*
 * sets specified value to specified option (by name) in option list
 * <list> by calling setter
 *
 * returns 0 on success, -1 otherwise
 */
int option_set(option_list_t* list, const char* name, const char* value) {
  while (list && strcmp(list->name, name)) {
    list = list->next;
  }

  if (list) {
    return list->setter(list->object, value);
  } else {
    return -1;
  }
}

/*
 * gets specified option (by name) from <list> by calling getter
 *
 * returns associated value, NULL on error
 */
const char* option_get(option_list_t* list, const char* name) {
  while (list && strcmp(list->name, name)) {
    list = list->next;
  }

  if (list) {
    return list->getter(list->object);
  } else {
    return NULL;
  }
}

/*
 * loads options from rc file
 *
 * returns 0 on success, -1 otherwise
 */
int option_restore_all(void) {
  return 0;
}

/*
 * saves options to rc file
 *
 * returns 0 on success, -1 otherwise
 */
int option_save_all(void) {
  return 0;
}

